/* no footer */
footer {
    display: none !important;
}

/* customize scrollbar in gr.Dataframe */
::-webkit-scrollbar {
    background: var(--background-fill-primary);
}
::-webkit-scrollbar-thumb {
    background-color: var(--border-color-primary);
    border: 4px solid transparent;
    border-radius: 100px;
    background-clip: content-box;
}
::-webkit-scrollbar-corner {
    background: var(--background-fill-primary);
}

.top-row {
    display: flex;
    align-items: center;
}

.top-row .image {
    display: flex;
    justify-content: center;
    align-items: center;
    background-color: transparent !important;
    border: none !important;
    box-shadow: none !important;
    padding: 0 !important;
    height: 150px;
}

.top-row .title-desc {
    justify-content: center;
    display: block;
    height: 100%; 
    background: var(--block-title-background-fill);
    border-radius: var(--radius-lg);
    font: var(--font);
    text-align: center;
    padding: var(--scale-0)
}

.top-row .title-desc .title {
    color: var(--block-title-text-color);
    font-size: var(--text-xxl);
    font-weight: var(--weight-extrabold);
}

.top-row .title-desc .desc {
    padding: 0px !important;
}

.search-container {
    display: flex;
    align-items: flex-end;
}


.search-container .button-not-running,
.search-container .button-running, 
.search-container .canceling {
    cursor: pointer;
    display: flex;
    justify-content: center;
    align-items: center;
    color: var(--color-grey-100);
    border-radius: var(--radius-md);
    font-size: var(--text-xxl);
    font-weight: var(--weight-extrabold);
}

.search-container .button-not-running {
    background: linear-gradient(to right, var(--color-pink-500), var(--button-primary-background-fill));
}
.search-container .button-running {
    background: var(--color-red-600);
}
.search-container .canceling {
    background: var(--color-grey-500);
}

.jobs-container {
    display: grid;
    grid-template-columns: repeat(auto-fill, minmax(var(--size-80), 1fr));
}
.card {
    justify-self: center;
    cursor: pointer;
    border: none var(--spacing-md) var(--block-title-border-color);
    border-radius: var(--radius-lg);
    width: var(--size-80);
    padding: var(--scale-2);
    font: var(--font);
    background: var(--block-label-background-fill);
}
.card:active {
    transform: scale(0.97);
}
.card .title {
    font-weight: var(--weight-bold);
    font-size: var(--text-lg);
    color: var(--block-title-text-color);
    margin-bottom: var(--scale-0);
}
.card .title .running{
    color: var(--color-red-600);
    font-size: var(--text-md);
}
.card .title .stopped{
    color: var(--color-grey-500);
    font-size: var(--text-md);
}
.card .title .completed{
    color: var(--color-green-700);
    font-size: var(--text-md);
}
.card .info { 
    color: var(--block-info-text-color);
    margin-bottom: var(--block-label-margin);
}

.details-container {
    display: flex;
    border-radius: 8px;
}
.details-container .title {
    font-weight: var(--weight-bold);
    font-size: var(--text-xl);
    padding-left: var(--scale-0);
    margin: var(--block-label-margin) 0 var(--block-label-margin) 0;
}
.details-container .info { 
    color: var(--block-info-text-color);
    padding-left: var(--scale-0);
    margin: var(--block-label-margin) 0 var(--block-label-margin) 0;
}

.details-container .items {
    margin-top: var(--scale-4);
}

.details-container .items .item {
    padding: var(--scale-0);
    margin-bottom: var(--scale-0);
    border: solid var(--block-label-border-width) var(--block-label-border-color);
}

.details-container .items .item .top{
    display: flex;
    align-items: center; 
}

.details-container .items .item .top .favicon{
    display: flex;
    align-items: center;
    margin-right: var(--scale-0);
}

.details-container .items .item .top .favicon img{
    height: 1.5em;
}

.details-container .items .item .top .title{
    display: flex;
    align-items: center; 
}

.details-container .items .item .bottom{
    display: flex;
    align-items: center; 
}

.details-container .items .item .bottom .ner.mp,
.details-container .items .item .bottom .ner.adr {
    padding: 0px 5px;
    border-radius: 4px;
    font-weight: var(--weight-bold);
}

.details-container .items .item .bottom .ner.mp {
    color: var(--color-grey-100);
    background-color: var(--color-purple-700);
} 

.details-container .items .item .bottom .ner.adr {
    background-color: var(--color-pink);
} 
